from __future__ import annotations

from collections.abc import Iterable


# Copied from textual codebase
def friendly_list(words: Iterable[str], joiner: str = "or", omit_empty: bool = True) -> str:
    """Generate a list of words as readable prose.

    >>> friendly_list(["foo", "bar", "baz"])
    "'foo', 'bar', or 'baz'"

    Args:
        words: A list of words.
        joiner: The last joiner word.

    Returns:
        List as prose.
    """
    words = [repr(word) for word in sorted(words, key=str.lower) if word or not omit_empty]
    if len(words) == 1:
        return words[0]
    elif len(words) == 2:
        word1, word2 = words
        return f"{word1} {joiner} {word2}"
    else:
        return f'{", ".join(words[:-1])}, {joiner} {words[-1]}'



from __future__ import annotations

import math
from dataclasses import dataclass

from rich.text import Text

from ._base import BaseHelp
from ._models import HelpEntry
from ._utils import arrange_items_in_columns


@dataclass
class Measurement:
    width: int
    entries: list[HelpEntry]
    separator: str

    col_padding: int = 1
    min_h: int = 3
    max_w: int = 32
    max_key_w: int = 10
    max_separator_w: int = 6

    def __post_init__(self) -> None:
        self._col_count: int | None = None
        self._col_w: int | None = None
        self._entry_key_w: int | None = None
        self._entry_sep_w: int | None = None
        self._entry_desc_w: int | None = None

    def _key_width(self) -> int:
        if self._entry_key_w is None:
            max_provided_key_width = max(len(entry.name) for entry in self.entries)
            self._entry_key_w = min(max_provided_key_width, self.max_key_w)
        return self._entry_key_w

    def _separator_width(self) -> int:
        if self._entry_sep_w is None:
            self._entry_sep_w = min(len(self.separator), self.max_separator_w)
        return self._entry_sep_w

    def _desc_width(self) -> int:
        if self._entry_desc_w is None:
            max_provided_desc_width = max(len(entry.description) for entry in self.entries)
            leftover_width = self.max_w - (self._key_width() + self._separator_width() + (2 * self.col_padding))
            self._entry_desc_w = min(max_provided_desc_width, leftover_width)
        return self._entry_desc_w

    @property
    def entry_widths(self) -> tuple[int, int, int]:
        return (self._key_width(), self._separator_width(), self._desc_width())

    @property
    def col_width(self) -> int:
        if self._col_w is None:
            entry_w = self._key_width() + self._separator_width() + self._desc_width() + (2 * self.col_padding)
            self._col_w = min(entry_w, self.max_w)
        return self._col_w

    @property
    def col_count(self) -> int:
        if self._col_count is None:
            self._col_count = max(1, math.floor(self.width / self.col_width))
        return self._col_count

    @property
    def height(self) -> int:
        expected_h = math.ceil(len(self.entries) / self.col_count)
        return max(expected_h, self.min_h)

    @staticmethod
    def default() -> Measurement:
        return Measurement(0, [], " ")


class LongHelp(BaseHelp):
    """A multi-line help widget to display key bindings"""

    COMPONENT_CLASSES = {
        "longhelp--separator",
        "longhelp--key",
        "longhelp--description",
    }

    DEFAULT_CSS = """
    LongHelp {
        width: 100w;
        height: auto;
        padding: 1;
    }
    LongHelp .longhelp--separator {
        color: #3C3C3C;
    }
    LongHelp .longhelp--key {
        color: #626262;
    }
    LongHelp .longhelp--description {
        color: #4A4A4A;
    }
    """

    def __init__(self, id: str, separator: str = "  ", debug: bool = False) -> None:
        super().__init__(id=id)
        self._separator = separator
        self._debug = debug

        self._measurement: Measurement = Measurement.default()
        self._pad_char = "#" if debug else " "
        self._pad_box = "|"

    def __entry_text(self, entry: HelpEntry) -> Text:
        key_style = self.get_component_rich_style("longhelp--key")
        description_style = self.get_component_rich_style("longhelp--description")
        separator_style = self.get_component_rich_style("longhelp--separator")

        (key_w, _, desc_w) = self._measurement.entry_widths

        line_text = Text(
            style=self.rich_style,
            no_wrap=True,
            overflow="ellipsis",
            end="",
        )

        line_text.pad(self._measurement.col_padding, self._pad_char)

        key_text = Text(entry.name.lower(), key_style, justify="right", overflow="ellipsis", no_wrap=True, end="")
        key_text.truncate(key_w, overflow="ellipsis")
        key_text.pad_left(key_w - len(entry.name))
        if self._debug:
            key_text.pad(1, self._pad_box)
        line_text.append(key_text)

        separator_text = Text(self._separator, separator_style, justify="center", overflow="crop", no_wrap=True, end="")
        line_text.append(separator_text)

        desc_text = Text(
            entry.description.lower(), description_style, justify="left", overflow="ellipsis", no_wrap=True, end=""
        )
        desc_text.truncate(desc_w, overflow="ellipsis")
        desc_text.pad_right(desc_w - len(entry.description))
        if self._debug:
            desc_text.pad(1, self._pad_box)
        line_text.append(desc_text)

        return line_text

    def _create_help(self) -> Text:
        # avoid division by zero
        if self.content_size.width == 0:
            return Text("")

        entries = self._get_entries_from_bindings()
        self._measurement = Measurement(self.content_size.width, entries, self._separator)
        entry_texts = [self.__entry_text(e) for e in entries]

        line_text = Text(
            style=self.rich_style,
            no_wrap=True,
            overflow="ellipsis",
            justify="left",
            end="",
        )

        line_separator = Text(
            "\n",
            style=self.rich_style,
            no_wrap=True,
            overflow="ellipsis",
            justify="left",
            end="",
        )

        table = arrange_items_in_columns(entry_texts, self._measurement.height, self._measurement.col_count)
        line_items = [line_text.join(rows) for rows in table]
        return line_separator.join(line_items)

    def get_content_height(self, *_) -> int:
        content = self._create_help()
        line_count = content.plain.count("\n") + 1
        return line_count



from __future__ import annotations

from typing import Literal

from rich.align import AlignMethod
from rich.cells import cell_len
from rich.console import Console, ConsoleOptions, ConsoleRenderable, RenderResult
from rich.style import Style
from rich.text import Text

LabelType = Literal["simple", "box"]
"""The names of the valid label types.

These are the types that can be used with a [`DividerWithLabel`][feathers.renderables.DividerWithLabel].
"""


class DividerWithLabel(ConsoleRenderable):
    def __init__(
        self,
        label: str,
        *,
        label_style: Style | None = None,
        label_padding: int = 2,
        label_type: LabelType = "simple",
        line_style: Style | None = None,
        width: int | None = None,
        align: AlignMethod = "left",
        end: str = "\n",
    ) -> None:
        """Create a DividerWithLabel renderable.

        Args:
            label: The text for label
            label_style: Style for the label
            label_padding: The left or right padding for label. Ignored if align in center. Defaults to 2
            label_type: The type of the label. Defaults to "simple"
            line_style: Style for the line
            width: The width of the divider. Defaults to console_options.max_width
            align: The alignment of the label. Defaults to "left"
        """

        self.label = label
        self.label_style = label_style
        self.label_padding = label_padding
        self._label_type = label_type
        self.line_style = line_style
        self.width = width
        self.align = align
        self.end = end

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        if self._label_type == "box":
            return self._render_box(console, options)
        else:
            return self._render_simple(console, options)

    def _render_simple(self, console: Console, options: ConsoleOptions) -> RenderResult:
        content_width = self.width or options.max_width
        spacing_around_prompt = 1
        prompt_width = cell_len(self.label) + (spacing_around_prompt * 2)
        label_style = self.label_style or console.style or Style()
        line_style = self.line_style or console.style or Style()

        left_divider_width = right_divider_width = 0
        if self.align == "left":
            left_divider_width = self.label_padding
            right_divider_width = content_width - prompt_width - left_divider_width
        elif self.align == "right":
            right_divider_width = self.label_padding
            left_divider_width = content_width - prompt_width - right_divider_width
        else:
            left_divider_width = right_divider_width = (content_width // 2) - (prompt_width // 2)

        yield Text("─" * left_divider_width, style=line_style, end="")
        yield Text(f'{" " * spacing_around_prompt}{self.label}{" " * spacing_around_prompt}', style=label_style, end="")
        yield Text("─" * right_divider_width, style=line_style, end=self.end)

    def _render_box(self, console: Console, options: ConsoleOptions) -> RenderResult:
        raise NotImplementedError("Not yet implemented")



from __future__ import annotations

from abc import abstractmethod

from rich.text import Text
from textual.widget import Widget, events
from textual.widgets import Static

from ._models import HelpEntry


class BaseHelp(Static):
    # called by textual
    def _on_resize(self, _: events.Resize) -> None:
        self.__reset(None)

    # called by textual
    def _on_mount(self, _: events.Mount) -> None:
        self.watch(self.screen, "focused", self.__reset)
        self.watch(self.screen, "stack_updates", self.__reset)
        self.__reset(None)

    # called by textual
    def _notify_style_update(self) -> None:
        self.__reset(None)

    @abstractmethod
    def _create_help(self) -> Text:
        pass

    def _can_switch(self) -> bool:
        return True

    def _get_entries_from_bindings(self) -> list[HelpEntry]:
        binding_entries = set([HelpEntry.from_binding(b) for (_, b) in self.app.namespace_bindings.values() if b.show])
        return sorted(list(binding_entries), key=lambda entry: entry.name)

    def __reset(self, _: Widget | None) -> None:
        self.update(self._create_help())



from __future__ import annotations

from enum import Enum
from typing import Literal

from feathers.utils import friendly_list
from feathers.widgets import CachedView

from ._models import ChatEntry, ChatRenderable, ChatRenderer
from ._renderers import MarkdownChatRenderer, MinimalChatRenderer

ChatStyle = Literal["minimal"]
"""The names of the supported chat styles.

These are the styles that can be used with a [`Chat`][feathers.widgets._chat].
"""

_VALID_CHAT_STYLES = {"minimal"}


class InvalidChatStyle(Exception):
    """Exception raised if an invalid chat style is used."""


class RendererType(Enum):
    MINIMAL = "minimal"
    MARKDOWN = "markdown"


class Chat(CachedView):
    COMPONENT_CLASSES = {
        "chat--divider",
    }

    DEFAULT_CSS = """
    Chat {
        content-align: center middle;
        padding-left: 2;
    }

    Chat .chat--divider {
        color: $surface-lighten-2;
    }
    """

    def __init__(
        self,
        renderer: ChatRenderer | RendererType = RendererType.MINIMAL,
        *,
        max_width: int = 100,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(wrap=True, max_width=max_width, name=name, id=id, classes=classes, disabled=disabled)
        self._entries: list[ChatEntry] = []
        if isinstance(renderer, RendererType):
            if renderer == RendererType.MINIMAL:
                renderer = MinimalChatRenderer(self, max_width=self.max_width)
            elif renderer == RendererType.MARKDOWN:
                renderer = MarkdownChatRenderer(self, max_width=self.max_width)
            else:
                raise InvalidChatStyle(f"Valid chat styles are {friendly_list(_VALID_CHAT_STYLES)}")
        self._renderer = renderer

    def validate_chat_style(self, chat_style: str) -> str:
        """Validate the chat style."""
        if chat_style not in _VALID_CHAT_STYLES:
            raise InvalidChatStyle(f"Valid chat styles are {friendly_list(_VALID_CHAT_STYLES)}")
        return chat_style

    def add_chat(self, entry: ChatEntry) -> None:
        self._entries.append(entry)
        renderable = ChatRenderable(self._renderer, entry)
        self.add_entry(renderable)

    def get_entries(self) -> list[ChatEntry]:
        return self._entries

    def clear_all_chats(self) -> None:
        self._entries.clear()
        self.clear()



from __future__ import annotations

from abc import abstractmethod
from typing import ClassVar

from rich.repr import Result
from rich.style import Style
from textual.binding import Binding, BindingType
from textual.events import Blur, Focus
from textual.geometry import Offset
from textual.message import Message
from textual.reactive import reactive
from textual.scroll_view import ScrollView
from textual.strip import Strip


class NavigableView(ScrollView, can_focus=True):
    """A view which can be navigated using a cursor.

    This should be used as a base class of other scrollable views.

    Attributes:
        disable_cursor: Disable the cursor. Defaults to False
    """

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("left, h", "cursor_left", "move left", show=True),
        Binding("up, k", "cursor_up", "move up", show=True),
        Binding("right, l", "cursor_right", "move right", show=True),
        Binding("down, j", "cursor_down", "move down", show=True),
    ]
    """
    | Key(s) | Description |
    | :- | :- |
    | left, h | Move the cursor left. |
    | up, k | Move the cursor up. |
    | right, l | Move the cursor right. |
    | down, j | Move the cursor down. |
    """

    COMPONENT_CLASSES: ClassVar[set[str]] = {"navigation-box--cursor"}
    """
    | Class | Description |
    | :- | :- |
    | `navigation-box--cursor` | Target the cursor. |
    """

    DEFAULT_CSS = """
    .navigation-box--cursor {
        background: $surface;
        color: $text;
        text-style: reverse;
    }
    """

    class Changed(Message, bubble=False):
        """Posted when the cursor value changes.

        Can be handled using `on_cursor_view_changed` in a subclass of `CursorView` or in a parent
        widget in the DOM.

        Attributes:
            value: The value that the cursor was changed to. It is the offset of the cursor
            cursor_view: Reference to the originating parent CursorView
        """

        def __init__(self, cursor_view: NavigableView, value: Offset) -> None:
            super().__init__()
            self.cursor_view: NavigableView = cursor_view
            self.value: Offset = value
            self._content_size = self.cursor_view.content_size

        @property
        def control(self) -> NavigableView:  # type: ignore
            """Alias for self.cursor_view."""
            return self.cursor_view

    class SeekBottom(Message, bubble=False):
        """Posted when the cursor is pushed beyond bottom of the available content height.

        Can be handled using `on_cursor_view_seek_bottom` in a subclass of `CursorView` or in a parent
        widget in the DOM.

        Attributes:
            cursor_view: Reference to the originating parent CursorView
        """

        def __init__(self, cursor_view: NavigableView) -> None:
            super().__init__()
            self.cursor_view: NavigableView = cursor_view

    class SeekTop(Message, bubble=False):
        """Posted when the cursor is pushed beyond top of the availalbe content height.

        Can be handled using `on_cursor_view_seek_top` in a subclass of `CursorView` or in a parent
        widget in the DOM.

        Attributes:
            cursor_view: Reference to the originating parent CursorView
        """

        def __init__(self, cursor_view: NavigableView) -> None:
            super().__init__()
            self.cursor_view: NavigableView = cursor_view

    class SeekLeft(Message, bubble=False):
        """Posted when the cursor is pushed beyond left of the availalbe content width.

        Can be handled using `on_cursor_view_seek_left` in a subclass of `CursorView` or in a parent
        widget in the DOM.

        Attributes:
            cursor_view: Reference to the originating parent CursorView
        """

        def __init__(self, cursor_view: NavigableView) -> None:
            super().__init__()
            self.cursor_view: NavigableView = cursor_view

    class SeekRight(Message, bubble=False):
        """Posted when the cursor is pushed beyond right of the availalbe content width.

        Can be handled using `on_cursor_view_seek_right` in a subclass of `CursorView` or in a parent
        widget in the DOM.

        Attributes:
            cursor_view: Reference to the originating parent CursorView
        """

        def __init__(self, cursor_view: NavigableView) -> None:
            super().__init__()
            self.cursor_view: NavigableView = cursor_view

    cursor_position: reactive[Offset] = reactive(Offset(0, 0))
    _cursor_visible = reactive(False)

    def __init__(
        self,
        *,
        disable_cursor: bool = False,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        """Initialise the `Input` widget.

        Args:
            mode: A mode in which cursor view should operate.
            name: Optional name for the input widget.
            id: Optional ID for the widget.
            classes: Optional initial classes for the widget.
            disabled: Whether the input is disabled or not.
        """
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self.cursor_disabled = disable_cursor
        if disable_cursor:
            NavigableView.BINDINGS.clear()
            self.can_focus = False

    @abstractmethod
    def line_count(self) -> int:
        pass

    @abstractmethod
    def line_width(self, y: int) -> int:
        pass

    def add_cursor(self, y: int, line: Strip) -> Strip:
        if self.cursor_position.y != y or self._is_cursor_hidden:
            return line

        cursor_style = self.get_component_rich_style("navigation-box--cursor")

        cursor_pos = max(0, min(self.cursor_position.x, line.cell_length - 1))
        return self._add_style_to_strip(line, cursor_pos, cursor_style)

    def _add_style_to_strip(self, line: Strip, pos: int, style: Style) -> Strip:
        # for empty strips, we still want to show the cursor
        if line.cell_length == 0:
            line = Strip.blank(1)

        strips = line.divide([pos, pos + 1, line.cell_length])
        if len(strips) < 2:
            return line
        changed_strips = list(strips)
        cursor_strip = strips[1]
        changed_strips[1] = cursor_strip.apply_style(style)
        return Strip.join(changed_strips)

    @property
    def _is_cursor_hidden(self) -> bool:
        return self.cursor_disabled or not self._cursor_visible

    def watch_cursor_position(self, value: Offset) -> None:
        self.post_message(self.Changed(self, value))

    def validate_cursor_position(self, cursor_position: Offset) -> Offset:
        content_w, content_h = self.scrollable_content_region.size
        if cursor_position.y == content_h:
            self.post_message(self.SeekBottom(self))
        if cursor_position.y < 0:
            self.post_message(self.SeekTop(self))
        if cursor_position.x < 0:
            self.post_message(self.SeekLeft(self))
        if cursor_position.x == content_w:
            self.post_message(self.SeekRight(self))

        y = min(cursor_position.y, self.line_count() - 1, content_h - 1)
        line_width = self.line_width(y)
        x = min(cursor_position.x, line_width - 1, content_w - 1)
        pos = Offset(x, y)
        return pos.clamped

    def on_navigable_view_seek_bottom(self, _: NavigableView.SeekBottom) -> None:
        self.scroll_down(animate=False)

    def on_navigable_view_seek_top(self, _: NavigableView.SeekTop) -> None:
        self.scroll_up(animate=False)

    def on_navigable_view_seek_left(self, _: NavigableView.SeekBottom) -> None:
        self.scroll_left(animate=False)

    def on_navigable_view_seek_right(self, _: NavigableView.SeekTop) -> None:
        self.scroll_right(animate=False)

    def _on_blur(self, _: Blur) -> None:
        self._cursor_visible = False

    def _on_focus(self, _: Focus) -> None:
        self._cursor_visible = True

    def action_cursor_left(self) -> None:
        """Move the cursor one position to the left."""
        if not self.cursor_disabled:
            self.cursor_position += Offset(-1, 0)

    def action_cursor_right(self) -> None:
        """Move the cursor one position to the right."""
        if not self.cursor_disabled:
            self.cursor_position += Offset(1, 0)

    def action_cursor_up(self) -> None:
        """Move the cursor one position upwards."""
        if not self.cursor_disabled:
            self.cursor_position += Offset(0, -1)

    def action_cursor_down(self) -> None:
        """Move the cursor one position downwards."""
        if not self.cursor_disabled:
            self.cursor_position += Offset(0, 1)

    def __rich_repr__(self) -> Result:
        yield from super().__rich_repr__()
        yield "cursor_disabled", self.cursor_disabled
        yield "cursor_position", self.cursor_position
        yield "_cursor_visible", self._cursor_visible



from ._cached_view import CachedView
from ._nav_view import NavigableView
from .help import Help, HelpEntry, HelpProvider

__all__ = ["Help", "HelpEntry", "HelpProvider", "NavigableView", "CachedView"]



from __future__ import annotations

from typing import TypeVar

T = TypeVar("T")


def arrange_items_in_columns(items: list[T], rows: int, cols: int) -> list[tuple[T, ...]]:
    """
    Distribute items from a list into a table-like structure with the specified number of rows and columns.

    Items are assigned to columns first, then to rows. The output is a list of tuples, where each
    inner tuple represents a row in the table-like structure.

    Args:
        items (list): The input list of items to distribute.
        rows (int): The number of rows in the resulting table-like structure.
        cols (int): The number of columns in the resulting table-like structure.

    Returns:
        List[tuple[T]]: A list of tuples, where each inner list represents a row in the table-like structure.

    Example:
        >>> items = ['a', 'b', 'c', 'd', 'e', 'f', 'g']
        >>> rows = 5
        >>> cols = 3
        >>> distribute_items_in_columns(items, rows, cols)
        [['a', 'f'], ['b', 'g'], ['c'], ['d'], ['e']]
    """
    result = []
    for i in range(rows):
        row_items = []
        for j in range(cols):
            index = i + j * rows
            if index < len(items):
                row_items.append(items[index])
        result.append(tuple(row_items))
    return result



from __future__ import annotations

from typing import cast

from rich.text import Text

from ._base import BaseHelp
from ._models import HelpEntry, HelpProvider


class ShortHelp(BaseHelp):
    """A single line help widget to display key bindings"""

    COMPONENT_CLASSES = {
        "shorthelp--separator",
        "shorthelp--key",
        "shorthelp--description",
    }

    DEFAULT_CSS = """
    ShortHelp {
        padding: 0 1;
        width: 100w;
        height: 1;
    }
    ShortHelp .shorthelp--separator {
        color: #3C3C3C
    }
    ShortHelp .shorthelp--key {
        color: #626262
    }
    ShortHelp .shorthelp--description {
        color: #4A4A4A
    }
    """

    def __init__(
        self,
        id: str,
        toggle_key: HelpEntry,
        separator: str,
        placeholder_text: str = "No key mappings found. Trying changing focus",
    ) -> None:
        super().__init__(id=id)
        self.toggle_key = toggle_key
        self.separator = separator
        self.placeholder_text: str = placeholder_text

        self._can_switch_to_long_help: bool = False
        self._text_style = Text(
            style=self.rich_style,
            no_wrap=True,
            overflow="ellipsis",
            justify="left",
            end="",
        )

    def _can_switch(self) -> bool:
        return self._can_switch_to_long_help

    def _create_help(self) -> Text:
        text = self._text_style.copy()
        entries = self.__get_entries()

        if self.__has_long_help_entries(entries):
            self._can_switch_to_long_help = True
            entries.append(self.toggle_key)
        else:
            self._can_switch_to_long_help = False

        if len(entries) == 0:
            return self.__get_placeholder_text()

        separator_style = self.get_component_rich_style("shorthelp--separator")
        separator = text.append(self.separator, separator_style)

        entry_texts = [self.__entry_text(e) for e in entries]
        return separator.join(entry_texts)

    def __entry_text(self, entry: HelpEntry) -> Text:
        key_style = self.get_component_rich_style("shorthelp--key")
        description_style = self.get_component_rich_style("shorthelp--description")
        return Text.assemble((entry.name.lower(), key_style), (f" {entry.description.lower()}", description_style))

    def __get_placeholder_text(self) -> Text:
        description_style = self.get_component_rich_style("shorthelp--description")
        return Text(
            self.placeholder_text,
            style=description_style,
            no_wrap=True,
            overflow="ellipsis",
            justify="left",
            end="",
        )

    def __get_entries(self) -> list[HelpEntry]:
        focused_widget = self.app.focused

        if focused_widget is None:
            self.log.info("No focused widget found")
            return []

        if isinstance(focused_widget, HelpProvider):
            return focused_widget.short_help_keys()

        if not hasattr(focused_widget, "short_help_keys") or not callable(getattr(focused_widget, "short_help_keys")):
            self.log.info("No short_help_keys method found on the current widget")
            return []

        widget = cast(HelpProvider, focused_widget)  # make mypy happy
        keys = widget.short_help_keys()

        if not isinstance(keys, list):
            self.log.warning("short_help_keys should return a list")
            return []

        if not all(isinstance(key, HelpEntry) for key in keys):
            self.log.warning("Not all items returned by short_help_keys are of they HelpEntry")
            return []

        return list(dict.fromkeys(keys))  ## dedupe list

    def __has_long_help_entries(self, entries: list[HelpEntry]) -> bool:
        long_help_entries = self._get_entries_from_bindings()
        if self.toggle_key in long_help_entries:
            long_help_entries.remove(self.toggle_key)
        return bool(set(long_help_entries) - set(entries))



from __future__ import annotations

from typing import cast

from rich.console import RenderableType
from rich.highlighter import ReprHighlighter
from rich.pretty import Pretty
from rich.protocol import is_renderable
from rich.text import Text
from textual.geometry import Size
from textual.strip import Strip

from feathers.cache import CacheListener, RenderablesCache, RenderableWithOptions

from ._nav_view import NavigableView


class CachedView(NavigableView, CacheListener):
    # scrollbar-gutter here is a fix to calculate scrollbar_gutter which impact the scrollable_content_region
    # calculation in `on_resize`. This looks like a bug in base widget code. Without this, the horizontal scrollbar
    # shows up for few characters on first call to `on_resize`
    DEFAULT_CSS = """
    CachedView{
        scrollbar-gutter: stable;
    }
    """

    max_width: int | None = None
    wrap: bool = False
    highlight: bool = False
    markup: bool = False
    auto_scroll: bool = True

    def __init__(
        self,
        *,
        max_width: int | None = None,
        wrap: bool = False,
        highlight: bool = False,
        markup: bool = False,
        auto_scroll: bool = True,
        enable_cursor: bool = False,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        """Create a TextLog widget.

        Args:
            wrap: Enable word wrapping (default is off).
            highlight: Automatically highlight content.
            markup: Apply Rich console markup.
            auto_scroll: Enable automatic scrolling to end.
            enable_cursor: Enable cursor which name this view navigable. This also make this view focusable.
            name: The name of the text log.
            id: The ID of the text log in the DOM.
            classes: The CSS classes of the text log.
            disabled: Whether the text log is disabled or not.
        """
        super().__init__(disable_cursor=not enable_cursor, name=name, id=id, classes=classes, disabled=disabled)
        self.max_width = max_width
        """Enforce max width of all the content"""
        self.wrap = wrap
        """Enable word wrapping."""
        self.highlight = highlight
        """Automatically highlight content."""
        self.markup = markup
        """Apply Rich console markup."""
        self.auto_scroll = auto_scroll
        """Automatically scroll to the end on write."""
        self.highlighter = ReprHighlighter()

        self._renderables_cache: RenderablesCache = RenderablesCache(self.app.console, listener=self)

    def notify_style_update(self) -> None:
        self._renderables_cache.refresh()

    def _extract_renderable(
        self,
        content: RenderableType | object,
        id: str | None,
        width: int | None = None,
        expand: bool | None = None,
        shrink: bool | None = None,
    ) -> RenderableWithOptions:
        renderable: RenderableType
        if not is_renderable(content):
            renderable = Pretty(content)
        else:
            if isinstance(content, str):
                if self.markup:
                    renderable = Text.from_markup(content)
                else:
                    renderable = Text(content)
                if self.highlighter is not None:
                    renderable = self.highlighter(renderable)
            else:
                renderable = cast(RenderableType, content)

        return RenderableWithOptions(renderable, id, width, expand, shrink, self.wrap)

    def add_entry(
        self,
        content: RenderableType | object,
        id: str | None = None,
        *,
        width: int | None = None,
        expand: bool = False,
        shrink: bool = True,
        scroll_end: bool | None = None,
    ) -> CachedView:
        """Write text or a rich renderable.

        Args:
            content: Rich renderable (or text).
            id: The renderable id which can later be used to remove or update the renderable. Can be missing if no
            update or remove is intended
            width: Width to render or `None` to use optimal width. Only used if either or both expand and shrink are
            True.
            expand: Enable expand to widget width, or `False` to use `width`.
            shrink: Enable shrinking of content to fit width.
            scroll_end: Enable automatic scroll to end, or `None` to use `self.auto_scroll`.

        Returns:
            The `CachedView` instance.
        """

        width = width or self.max_width
        renderable = self._extract_renderable(content, id, width, expand, shrink)
        self._renderables_cache.add(renderable)

        auto_scroll = self.auto_scroll if scroll_end is None else scroll_end
        if auto_scroll:
            self.scroll_end(animate=False)

        return self

    def remove_entry(self, id: str) -> CachedView:
        """Remove a renderable from cache.

        Args:
            id: The id of the renderable to remove. If id is not found, the operation is ignored.

        Returns:
            The `CachedView` instance.
        """
        self._renderables_cache.remove(id)
        return self

    def clear(self) -> CachedView:
        self._renderables_cache.clear()
        self.max_width = 0
        self.virtual_size = Size(0, 0)
        self.refresh()
        return self

    def on_resize(self):
        self._renderables_cache.content_width = self.scrollable_content_region.width

    def on_cache_update(self):
        self.virtual_size = self._renderables_cache.virtual_size
        if self.auto_scroll:
            self.scroll_end(animate=False)

    def line_count(self) -> int:
        return len(self._renderables_cache)

    def line_width(self, y: int) -> int:
        _, scroll_y = self.scroll_offset
        line = self._renderables_cache.strip_at(scroll_y + y)
        if line is None:
            return 0
        return line.cell_length

    def render_line(self, y: int) -> Strip:
        scroll_x, scroll_y = self.scroll_offset
        line = self._render_line(scroll_y + y, scroll_x, self.size.width)
        strip = line.apply_style(self.rich_style)
        strip = self.add_cursor(y, line)
        return strip

    def _render_line(self, y: int, scroll_x: int, width: int) -> Strip:
        strip = self._renderables_cache.strip_at(y)
        if strip is None:
            return Strip.blank(width, self.rich_style)

        line = strip.crop(scroll_x, scroll_x + width)

        return line



from __future__ import annotations

from rich.console import Console, ConsoleOptions, ConsoleRenderable, RenderResult
from rich.markdown import Markdown
from rich.style import Style
from rich.text import Text
from textual.widget import Widget

from feathers.renderables import DividerWithLabel

from ._models import ChatEntry, ChatRenderer


class MinimalChatRenderer(ChatRenderer):
    def __init__(self, widget: Widget, *, max_width: int | None = None) -> None:
        super().__init__(widget)
        self._max_width = max_width
        self._blank_line = Text(style=self._widget.rich_style)

    def render(self, entry: ChatEntry, console: Console, options: ConsoleOptions) -> RenderResult:
        divider_style = self._widget.get_component_rich_style("chat--divider")
        content_width = self._max_width or self._widget.content_size.width
        divider = Text("─" * content_width, style=divider_style)

        yield self._blank_line
        yield Text.assemble(
            (f"{entry.participant.name}: ", Style(color=entry.participant.color)),
            (entry.message, Style(color="white")),
            end="\n",
        )
        yield self._blank_line
        yield divider


class MarkdownChatRenderer(ChatRenderer):
    def __init__(
        self, widget: Widget, prompt_renderable: ConsoleRenderable | None = None, *, max_width: int | None = None
    ) -> None:
        super().__init__(widget)
        self._max_width = max_width
        self._prompt_renderable = prompt_renderable
        self._blank_line = Text(style=self._widget.rich_style)

    def render(self, entry: ChatEntry, console: Console, options: ConsoleOptions) -> RenderResult:
        divider_style = self._widget.get_component_rich_style("chat--divider")
        prompt_style = Style(color=entry.participant.color)
        prompt_renderable = self._prompt_renderable or DividerWithLabel(
            entry.participant.name,
            label_style=prompt_style,
            line_style=divider_style,
            width=self._max_width,
            end="\n\n",
        )

        yield prompt_renderable
        yield Markdown(entry.message, code_theme="github-dark")
        yield self._blank_line



from __future__ import annotations

from abc import abstractmethod
from dataclasses import dataclass

from rich.console import Console, ConsoleOptions, RenderResult
from textual.widget import Widget


@dataclass
class Participant:
    name: str
    icon: str = ""
    color: str | None = None


@dataclass
class ChatEntry:
    message: str
    participant: Participant
    timestamp: float | None = None


class ChatRenderer:
    """An object that supports the console protocol and can render a chat"""

    def __init__(self, widget: Widget) -> None:
        self._widget = widget

    @abstractmethod
    def render(self, entry: ChatEntry, console: Console, options: ConsoleOptions) -> RenderResult:
        pass


class ChatRenderable:
    """An object that supports the console protocol and can render a chat"""

    def __init__(self, renderer: ChatRenderer, entry: ChatEntry) -> None:
        self._renderer = renderer
        self._entry = entry

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        return self._renderer.render(self._entry, console, options)



from __future__ import annotations

from typing import Literal

from textual.app import ComposeResult
from textual.message import Message
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import ContentSwitcher

from feathers.utils import friendly_list

from ._base import BaseHelp
from ._long_help import LongHelp
from ._models import HelpEntry
from ._short_help import ShortHelp

HelpMode = Literal["short", "long"]
"""The names of the valid help mode.

These are the modes that can be used with a [`Help`][feathers.widgets.Help].
"""

_VALID_HELP_MODES = {"short", "long"}


class InvalidHelpMode(Exception):
    """Exception raised if an invalid button variant is used."""


class Help(Widget):
    """A single and multi line mode help widget to display key bindings"""

    DEFAULT_CSS = """
    Help {
        dock: bottom;
        width: 100w;
        min-height: 1;
        height: auto;
    }

    Help > ContentSwitcher {
        height: auto;
    }
    """

    mode: reactive[HelpMode] = reactive[HelpMode]("short")
    """The current display mode, either short or long."""

    class Toggled(Message, bubble=True):
        """Posted when the help mode changes."""

        def __init__(self, help: Help, mode: HelpMode) -> None:
            """Initialise the message.

            Args:
                help: The help widget sending the message.
                mode: The current mode of the help button.
            """
            super().__init__()
            self._help = help
            """A reference to the help widget that was changed."""
            self.mode = mode
            """The value of the help mode after the change."""

    def __init__(
        self,
        toggle_key: HelpEntry = HelpEntry("?", "toggle help", is_primary=True),
        short_help_separator: str = " • ",
        short_help_placeholder: str = "No available key mappings",
        long_help_separator: str = "  ",
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)

        self._short_help: BaseHelp = ShortHelp("short-help", toggle_key, short_help_separator, short_help_placeholder)
        self._long_help: BaseHelp = LongHelp("long-help", long_help_separator)

    def compose(self) -> ComposeResult:
        with ContentSwitcher(initial="short-help"):
            yield self._short_help
            yield self._long_help

    def toggle_help(self) -> Help:
        """Toggle the help view between short and long.

        Returns:
            The `Help` instance.
        """
        self._switch_view()
        return self

    def action_toggle_help(self) -> None:
        """Toggle the mode of the help when called as an action.

        This would normally be used for a keyboard binding.
        """
        self.toggle_help()

    def validate_mode(self, mode: str) -> str:
        """Validate the mode."""
        if mode not in _VALID_HELP_MODES:
            raise InvalidHelpMode(f"Valid help modes are {friendly_list(_VALID_HELP_MODES)}")
        return mode

    def watch_mode(self) -> None:
        """React to the mode being changed."""

        self.post_message(self.Toggled(self, self.mode))

    def _switch_view(self) -> None:
        switcher = self.query_one(ContentSwitcher)
        if switcher.current == "short-help" and self._short_help._can_switch():
            switcher.current = "long-help"
            self.mode = "long"
        else:
            switcher.current = "short-help"
            self.mode = "short"



from __future__ import annotations

from collections import OrderedDict
from dataclasses import dataclass
from typing import NewType

from rich.console import Console, ConsoleOptions, RenderableType
from rich.measure import Measurement
from rich.segment import Segment
from rich.text import Text
from textual.geometry import Size
from textual.strip import Strip


class CacheListener:
    def on_cache_update(self):
        pass


CacheId = NewType("CacheId", str)


@dataclass
class RenderableWithOptions:
    renderableType: RenderableType
    id: str | None = None
    width: int | None = None
    expand: bool | None = False
    strink: bool | None = True
    wrap: bool = False


class RenderablesCache:
    """Represents a collection of renderables for a Textual Widget.

    A renderables can result in Strips which are cached. The renderables can
    be added and removed and Strips are cached for better performance.

    """

    def __init__(
        self, console: Console, options: ConsoleOptions | None = None, listener: CacheListener | None = None
    ) -> None:
        self._console = console
        self._options = options if options is not None else console.options
        self._listener = listener

        self._all_renderables: OrderedDict[CacheId, RenderableWithOptions] = OrderedDict()
        self._renderables_added: list[CacheId] = []
        self._renderables_removed: list[CacheId] = []
        self._cache: list[tuple[CacheId, Strip]] = []
        self._virtual_size: Size = Size(0, 0)

        self._content_width: int | None = None

    @property
    def virtual_size(self) -> Size:
        return self._virtual_size

    @property
    def content_width(self) -> int | None:
        return self._content_width

    @content_width.setter
    def content_width(self, value: int) -> None:
        if value != self._content_width:
            self._content_width = value
            self.refresh()

    def strip_at(self, index: int) -> Strip | None:
        """Get the strip at given index from cache"""
        if index >= len(self._cache):
            return None
        return self._cache[index][1]

    def add(self, renderable: RenderableWithOptions):
        """Add a new renderable. Pass the id in renderable if you intend to update or remove it later"""
        renderable_id = CacheId(renderable.id) if renderable.id else CacheId(str(id(renderable)))
        if renderable_id not in self._all_renderables:
            self._all_renderables[renderable_id] = renderable
            self._renderables_added.append(renderable_id)
            self._update_cache()

    def remove(self, id: str):
        """Remove the renderable from cache. If missing, the operation is ignored"""
        renderable_id = CacheId(id)
        if id in self._all_renderables:
            self._renderables_removed.append(renderable_id)
            self._all_renderables.pop(renderable_id)
            self._update_cache()

    def refresh(self) -> None:
        self._cache.clear()
        self._renderables_added.clear()
        self._renderables_removed.clear()
        for id in self._all_renderables.keys():
            self._renderables_added.append(id)
        self._virtual_size = Size(0, 0)
        self._update_cache()

    def clear(self) -> None:
        self._all_renderables.clear()
        self.refresh()

    def _update_cache(self) -> None:
        if not self._content_width:
            return None
        is_updated = False

        for id in self._renderables_added:
            self._add_to_cache(id, self._all_renderables[id])
            is_updated = True
        self._renderables_added.clear()

        for id in self._renderables_removed:
            is_updated = True
            self._remove_from_cache(id)
        self._renderables_removed.clear()

        if is_updated and self._listener is not None:
            self._listener.on_cache_update()

    def _add_to_cache(self, id: CacheId, renderable: RenderableWithOptions) -> None:
        strips = self._extract_lines(renderable)
        if strips is None:
            return
        for strip in strips:
            self._cache.append((id, strip))

    def _remove_from_cache(self, id: CacheId) -> None:
        # By iterating over the list in reverse order, we can safely remove items without changing
        # the indices of the remaining items in the list. This approach can be more memory-efficient
        # than making a copy of the list to iterate over, especially for large lists, since we only need
        # to keep track of the current index and don't need to create a copy of the list.
        for i in range(len(self._cache) - 1, -1, -1):
            if id in self._cache[i]:
                self._cache.pop(i)
        self._refresh_virtual_size()

    def _extract_lines(
        self,
        renderable: RenderableWithOptions,
    ) -> list[Strip] | None:
        renderableType, width, expand, shrink, wrap = (
            renderable.renderableType,
            renderable.width,
            renderable.expand,
            renderable.strink,
            renderable.wrap,
        )
        render_options = self._console.options

        if isinstance(renderable, Text) and not wrap:
            render_options = render_options.update(overflow="ignore", no_wrap=True)

        render_width = Measurement.get(self._console, render_options, renderableType).maximum
        optimal_width = self._content_width if width is None else width

        if optimal_width:
            if expand and render_width < optimal_width:
                render_width = optimal_width
            if shrink and render_width > optimal_width:
                render_width = optimal_width

        segments = self._console.render(renderableType, render_options.update_width(render_width))
        lines = list(Segment.split_lines(segments))
        if not lines:
            return None

        strips = Strip.from_lines(lines)
        for strip in strips:
            strip.adjust_cell_length(render_width)

        max_width = max(strip.cell_length for strip in strips)
        new_width = max(self._virtual_size.width, max_width)
        new_height = self._virtual_size.height + len(strips)
        self._virtual_size = Size(new_width, new_height)
        return strips

    def _refresh_virtual_size(self):
        max_width = max(strip[1].cell_length for strip in self._cache)
        self._virtual_size = Size(max_width, len(self._cache))

    def __len__(self) -> int:
        return len(self._cache)
